# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.airflow.utils import get_job_name
from openlineage.client.facet import SqlJobFacet
from openlineage.common.provider.redshift_data import RedshiftDataDatasetsProvider
from openlineage.common.sql import SqlMeta, parse


class RedshiftDataExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["RedshiftDataOperator"]

    def extract(self) -> Optional[TaskMetadata]:
        return None

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        self.log.debug(f"extract_on_complete({task_instance})")
        job_facets = {"sql": SqlJobFacet(self.operator.sql)}

        self.log.debug(f"Sending SQL to parser: {self.operator.sql}")
        sql_meta: Optional[SqlMeta] = parse(
            self.operator.sql,
            dialect=self.dialect,
            default_schema=self.default_schema
        )
        self.log.debug(f"Got meta {sql_meta}")
        try:
            redshift_job_id = self._get_xcom_redshift_job_id(task_instance)
            if redshift_job_id is None:
                raise Exception(
                    "Xcom could not resolve Redshift job id. Job may have failed."
                )
        except Exception as e:
            self.log.error(f"Cannot retrieve job details from {e}", exc_info=True)
            return TaskMetadata(
                name=get_job_name(task=self.operator),
                run_facets={},
                job_facets=job_facets,
            )

        client = self.operator.hook.conn

        redshift_details = [
            "database",
            "cluster_identifier",
            "db_user",
            "secret_arn",
            "region",
        ]

        connection_details = {
            detail: getattr(self.operator, detail) for detail in redshift_details
        }

        stats = RedshiftDataDatasetsProvider(
            client=client, connection_details=connection_details
        ).get_facets(
            job_id=redshift_job_id,
            inputs=sql_meta.in_tables if sql_meta else [],
            outputs=sql_meta.out_tables if sql_meta else [],
        )

        return TaskMetadata(
            name=get_job_name(task=self.operator),
            inputs=[ds.to_openlineage_dataset() for ds in stats.inputs],
            outputs=[ds.to_openlineage_dataset() for ds in stats.output],
            run_facets=stats.run_facets,
            job_facets={"sql": SqlJobFacet(self.operator.sql)},
        )

    @property
    def dialect(self):
        return "redshift"

    @property
    def default_schema(self):
        # TODO: check default schema in redshift
        return "public"

    def _get_xcom_redshift_job_id(self, task_instance):
        redshift_job_id = task_instance.xcom_pull(task_ids=task_instance.task_id)

        self.log.debug(f"redshift job id: {redshift_job_id}")
        return redshift_job_id



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import List, Optional

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.client.run import Dataset

log = logging.getLogger(__name__)


def generate_s3_dataset(path) -> Dataset:
    return Dataset(
        namespace="s3://{}".format(path.replace("s3://", "").split('/')[0]),
        name=path,
        facets={}
    )


class SageMakerProcessingExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ['SageMakerProcessingOperator', 'SageMakerProcessingOperatorAsync']

    def extract(self) -> Optional[TaskMetadata]:
        pass

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:

        xcom_values = task_instance.xcom_pull(task_ids=task_instance.task_id)

        inputs = []
        outputs = []

        try:
            inputs, outputs = self._get_s3_datasets(
                processing_inputs=xcom_values['Processing']['ProcessingInputs'],
                processing_outputs=xcom_values['Processing']['ProcessingOutputConfig']['Outputs']
            )
        except KeyError as e:
            log.error(f"Could not find input/output information in Xcom. {e}")

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=inputs,
            outputs=outputs,
        )

    @staticmethod
    def _get_s3_datasets(processing_inputs, processing_outputs):

        inputs = []
        outputs = []

        try:
            for processing_input in processing_inputs:
                inputs.append(generate_s3_dataset(processing_input['S3Input']['S3Uri']))
        except Exception as e:
            log.error(f"Cannot find S3 input details. {e}", exc_info=True)

        try:
            for processing_output in processing_outputs:
                outputs.append(generate_s3_dataset(processing_output['S3Output']['S3Uri']))
        except Exception as e:
            log.error(f"Cannot find S3 output details. {e}", exc_info=True)

        return inputs, outputs


class SageMakerTransformExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ['SageMakerTransformOperator', 'SageMakerTransformOperatorAsync']

    def extract(self) -> Optional[TaskMetadata]:
        pass

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        log.debug(f"extract_on_complete({task_instance})")

        xcom_values = task_instance.xcom_pull(task_ids=task_instance.task_id)

        model_package_arn = None
        transform_input = None
        transform_output = None

        try:
            model_package_arn = xcom_values['Model']['PrimaryContainer']['ModelPackageName']
        except KeyError as e:
            log.error(f"Cannot find Model Package Name in Xcom values. {e}", exc_info=True)

        try:
            transform = xcom_values['Transform']
            transform_input = transform['TransformInput']['DataSource']['S3DataSource']['S3Uri']
            transform_output = transform['TransformOutput']['S3OutputPath']
        except KeyError as e:
            log.error(
                f"Cannot find some required input/output details in XCom. {e}", exc_info=True
            )

        inputs = []

        if transform_input is not None:
            inputs.append(generate_s3_dataset(transform_input))

        if model_package_arn is not None:
            model_data_urls = self._get_model_data_urls(model_package_arn)
            for model_data_url in model_data_urls:
                inputs.append(generate_s3_dataset(model_data_url))

        output = []
        if transform_output is not None:
            output.append(generate_s3_dataset(transform_output))

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=inputs,
            outputs=output
        )

    def _get_model_data_urls(self, model_package_arn):
        model_data_urls = []
        try:
            model_containers = self.operator.hook.get_conn().describe_model_package(
                ModelPackageName=model_package_arn
            )['InferenceSpecification']['Containers']

            for container in model_containers:
                model_data_urls.append(container['ModelDataUrl'])
        except Exception as e:
            log.error(f"Cannot retrieve model details. {e}", exc_info=True)

        return model_data_urls


class SageMakerTrainingExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ['SageMakerTrainingOperator', 'SageMakerTrainingOperatorAsync']

    def extract(self) -> Optional[TaskMetadata]:
        pass

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        log.debug(f"extract_on_complete({task_instance})")

        xcom_values = task_instance.xcom_pull(task_ids=task_instance.task_id)

        inputs = []
        output = []

        try:
            for input_data in xcom_values['Training']['InputDataConfig']:
                inputs.append(
                    generate_s3_dataset(input_data['DataSource']['S3DataSource']['S3Uri'])
                )
        except KeyError as e:
            log.error(f"Issues extracting inputs. {e}")

        try:
            output.append(
                generate_s3_dataset(xcom_values['Training']['ModelArtifacts']['S3ModelArtifacts'])
            )
        except KeyError as e:
            log.error(f"Issues extracting inputs. {e}")

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=inputs,
            outputs=output
        )



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import os

# Provide empty plugin for older version
from openlineage.airflow.macros import lineage_parent_id, lineage_run_id
from pkg_resources import parse_version

from airflow.plugins_manager import AirflowPlugin
from airflow.version import version as AIRFLOW_VERSION


def _is_disabled():
    return os.getenv("OPENLINEAGE_DISABLED", None) in [True, 'true', "True"]


if parse_version(AIRFLOW_VERSION) \
        < parse_version("2.3.0.dev0") or _is_disabled():      # type: ignore
    class OpenLineagePlugin(AirflowPlugin):
        name = "OpenLineagePlugin"
        macros = [lineage_run_id, lineage_parent_id]
else:
    from openlineage.airflow import listener

    # Provide entrypoint airflow plugin that registers listener module
    class OpenLineagePlugin(AirflowPlugin):     # type: ignore
        name = "OpenLineagePlugin"
        listeners = [listener]
        macros = [lineage_run_id, lineage_parent_id]



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0
#
# -*- coding: utf-8 -*-
import logging

from openlineage.airflow.version import __version__
from pkg_resources import parse_version

from airflow.version import version as AIRFLOW_VERSION

__author__ = """OpenLineage"""

if parse_version(AIRFLOW_VERSION) < parse_version("2.0.0"):     # type: ignore
    logging.warning(
        f'''
        OpenLineage support for Airflow version {AIRFLOW_VERSION} is REMOVED.
        Please make sure to upgrade your Airflow version to minimum of 2.0.0
        in order to continue using OpenLineage.
        '''
    )

from airflow.models import DAG

__all__ = ["DAG", "__version__"]



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import os
from typing import Optional, Type

from openlineage.airflow.extractors.base import BaseExtractor, DefaultExtractor
from openlineage.airflow.extractors.sql_check_extractors import get_check_extractors
from openlineage.airflow.extractors.sql_execute_query import (
    get_sql_execute_query_extractor,
    sql_extractors,
)
from openlineage.airflow.utils import import_from_string, try_import_from_string

_extractors = list(
    filter(
        lambda t: t is not None,
        [
            try_import_from_string(
                'openlineage.airflow.extractors.postgres_extractor.PostgresExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.mysql_extractor.MySqlExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.bigquery_extractor.BigQueryExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.great_expectations_extractor.GreatExpectationsExtractor'  # noqa: E501
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.snowflake_extractor.SnowflakeExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.python_extractor.PythonExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.bash_extractor.BashExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.redshift_sql_extractor.RedshiftSQLExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.redshift_data_extractor.RedshiftDataExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.trino_extractor.TrinoExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.athena_extractor.AthenaExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.sftp_extractor.SFTPExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.ftp_extractor.FTPExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.sagemaker_extractors.SageMakerProcessingExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.sagemaker_extractors.SageMakerTrainingExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.sagemaker_extractors.SageMakerTransformExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.s3_extractor.S3CopyObjectExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.s3_extractor.S3FileTransformExtractor'
            ),
            try_import_from_string(
                'openlineage.airflow.extractors.gcs_extractor.GCSToGCSExtractor'
            ),
        ],
    )
)

_extractors += get_check_extractors(
    try_import_from_string(
        'openlineage.airflow.extractors.sql_extractor.SqlExtractor'
    )
)

_check_providers = {
    "PostgresExtractor": "postgres",
    "MySqlExtractor": "mysql",
    "BigQueryExtractor": ["gcpbigquery", "google_cloud_platform"],
    "SnowflakeExtractor": "snowflake",
    "TrinoExtractor": "trino",
    "AthenaExtractor": "aws",
    "SFTPExtractor": ["sftp", "ssh"],
    "FTPExtractor": "ftp",
}


class Extractors:
    """
    This exposes implemented extractors, while hiding ones that require additional, unmet
    dependency. Patchers are a category of extractor that needs to hook up to operator's
    internals during DAG creation.
    """

    def __init__(self):
        # Do not expose extractors relying on external dependencies that are not installed
        self.extractors = {}
        self.default_extractor = DefaultExtractor

        for extractor in _extractors:
            for operator_class in extractor.get_operator_classnames():
                self.extractors[operator_class] = extractor

        # Comma-separated extractors in OPENLINEAGE_EXTRACTORS variable.
        # Extractors should implement BaseExtractor
        env_extractors = os.getenv("OPENLINEAGE_EXTRACTORS")
        if env_extractors is not None:
            for extractor in env_extractors.split(';'):
                extractor = import_from_string(extractor.strip())
                for operator_class in extractor.get_operator_classnames():
                    self.extractors[operator_class] = extractor

        # Previous way of adding extractors
        # Adding operator: extractor pairs registered via environmental variable in pattern
        # OPENLINEAGE_EXTRACTOR_<operator>=<path.to.ExtractorClass>
        # The value in extractor map is extractor class type - it needs to be instantiated.
        # We import the module provided and get type using importlib then.
        for key, value in os.environ.items():
            if key.startswith("OPENLINEAGE_EXTRACTOR_"):
                extractor = import_from_string(value)
                for operator_class in extractor.get_operator_classnames():
                    self.extractors[operator_class] = extractor

    def add_extractor(self, operator: str, extractor: Type):
        self.extractors[operator] = extractor

    def get_extractor_class(self, clazz: Type) -> Optional[Type[BaseExtractor]]:
        name = clazz.__name__
        if name in self.extractors:
            return self.extractors[name]

        def method_exists(method_name):
            method = getattr(clazz, method_name, None)
            if method:
                return callable(method)

        if method_exists("get_openlineage_facets_on_start") or method_exists(
            "get_openlineage_facets_on_complete"
        ):
            return self.default_extractor
        return None

    def instantiate_abstract_extractors(self, task) -> None:
        # instantiate sql check extractors
        from airflow.hooks.base import BaseHook

        if task.__class__.__name__ in (
            "SQLCheckOperator", "SQLValueCheckOperator", "SQLThresholdCheckOperator",
            "SQLIntervalCheckOperator", "SQLColumnCheckOperator", "SQLTableCheckOperator",
            "BigQueryTableCheckOperator", "BigQueryColumnCheckOperator"
        ):
            for extractor in self.extractors.values():
                conn_type = _check_providers.get(extractor.__name__, "")
                task_conn_type = None
                if hasattr(task, "gcp_conn_id"):
                    task_conn_type = BaseHook.get_connection(task.gcp_conn_id).conn_type
                elif hasattr(task, "conn_id"):
                    task_conn_type = BaseHook.get_connection(task.conn_id).conn_type
                if task_conn_type in conn_type:
                    check_extractors = get_check_extractors(extractor)
                    for check_extractor in check_extractors:
                        for operator_class in check_extractor.get_operator_classnames():
                            self.extractors[operator_class] = check_extractor
                    else:
                        return
            else:
                raise ValueError(
                    "Extractor for the given task's conn_type (%s) does not exist.",
                    task_conn_type
                )

        # instantiate sql execute query extractor
        if task.__class__.__name__ == "SQLExecuteQueryOperator":
            task_conn_type = BaseHook.get_connection(task.conn_id).conn_type
            extractor_name = sql_extractors.get(task_conn_type, None)
            extractor = list(
                filter(lambda x: x.__name__ == extractor_name, self.extractors.values())
            )
            if extractor:
                self.extractors[
                    "SQLExecuteQueryOperator"
                ] = get_sql_execute_query_extractor(extractor[0])
        return



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.airflow.extractors.extractors import Extractors
from openlineage.airflow.extractors.manager import ExtractorManager

__all__ = [Extractors, BaseExtractor, TaskMetadata, ExtractorManager]



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import os
import time
import uuid
from typing import Optional

from pkg_resources import parse_version

from airflow.lineage.backend import LineageBackend
from airflow.version import version as AIRFLOW_VERSION


class Backend:
    def __init__(self):
        from openlineage.airflow.adapter import OpenLineageAdapter
        from openlineage.airflow.extractors.manager import ExtractorManager
        self.extractor_manager = ExtractorManager()
        self.adapter = OpenLineageAdapter()
    """
    Send OpenLineage events to lineage backend via airflow's LineageBackend mechanism.
    The start and complete events are send when task instance completes.
    """

    def send_lineage(
            self,
            operator=None,
            inlets=None,
            outlets=None,
            context=None
    ):
        """
        Send_lineage ignores manually provided inlets and outlets. The data collection mechanism
        is automatic, and bases on the passed context.
        """
        from openlineage.airflow.adapter import OpenLineageAdapter
        from openlineage.airflow.utils import (
            DagUtils,
            get_airflow_run_facet,
            get_custom_facets,
            get_dagrun_start_end,
            get_job_name,
            get_task_location,
        )
        dag = context['dag']
        dagrun = context['dag_run']
        task_instance = context['task_instance']
        dag_run_id = self.adapter.build_dag_run_id(dag.dag_id, dagrun.run_id)

        run_id = str(uuid.uuid4())
        job_name = get_job_name(operator)

        task_metadata = self.extractor_manager.extract_metadata(
            dagrun=dagrun,
            task=operator,
            complete=True,
            task_instance=task_instance
        )

        task_uuid = OpenLineageAdapter.build_task_instance_run_id(
            operator.task_id, task_instance.execution_date, task_instance.try_number
        )
        start, end = get_dagrun_start_end(dagrun, dag)

        self.adapter.start_task(
            run_id=run_id,
            job_name=job_name,
            job_description=dag.description,
            event_time=DagUtils.get_start_time(task_instance.start_date),
            parent_job_name=dag.dag_id,
            parent_run_id=dag_run_id,
            code_location=get_task_location(operator),
            nominal_start_time=DagUtils.get_start_time(start),
            nominal_end_time=DagUtils.to_iso_8601(end),
            owners=dag.owner.split(', '),
            task=task_metadata,
            run_facets={
                **task_metadata.run_facets,
                **get_custom_facets(dagrun, operator, dagrun.external_trigger),
                **get_airflow_run_facet(dagrun, dag, task_instance, operator, task_uuid)
            }
        )

        self.adapter.complete_task(
            run_id=run_id,
            job_name=job_name,
            end_time=DagUtils.to_iso_8601(self._now_ms()),
            task=task_metadata,
        )

    @staticmethod
    def _now_ms():
        return int(round(time.time() * 1000))


class OpenLineageBackend(LineageBackend):
    # Constructing instance on first use and delegate calls to it - backwards compatible with 1.10
    backend: Optional[Backend] = None

    @classmethod
    def send_lineage(cls, *args, **kwargs):
        # Do not use LineageBackend approach when we can use plugins
        if parse_version(AIRFLOW_VERSION) >= parse_version("2.3.0.dev0"):
            return
        # Make this method a noop if OPENLINEAGE_DISABLED is set to true
        if os.getenv("OPENLINEAGE_DISABLED", None) in [True, 'true', "True"]:
            return
        if not cls.backend:
            cls.backend = Backend()
        return cls.backend.send_lineage(*args, **kwargs)
